// sys.h
// header per l' interfaccia offerta dal nucleo e dal modulo di IO
//

#ifndef _SYS_H_
#define _SYS_H_

// Prototipi delle chiamate di sistema disponibili
//

// Crea un nuovo processo, avente corpo F. A e` il parametro di F, PRIO la
//  priorita`, LIV specifica lo stato (utente o sistema), in ID viene ritornato
//  l' identificatore del nuovo processo e RISU vale FALSE in caso di errore,
//  true in caso di successo
//
extern "C" void activate_p(void f(int), int a, int prio, char liv,
		short &id, bool &risu);

// Termina il processo in esecuzione
//
extern "C" void terminate_p();

// Chiamata da main(), trasforma il programma nel processo a piu` bassa
//  priorita`, lo inserisce nella coda dei pronti, seleziona il processo a
//  priorita` maggiore e lo mette in esecuzione
//
extern "C" void begin_p();

// Ritorna il valore della variabile di nucleo processi, contenente il numero di
//  processi creati mediante activate_p e non terminati con terminate_p
//
extern "C" void give_num(int &lav);

// Cerca un descrittore di semaforo non utilizzato e lo inizializza a VAL.
//  L' indice del semaforo e` ritornato in INDEX_DES_S
//
extern "C" void sem_ini(int &index_des_s, int val, bool &risu);

// Decrementa il valore del semaforo di cui SEM e` l' indice; se il nuovo
//  valore e` negativo blocca il processo su SEM
//
extern "C" void sem_wait(int sem);

// Incrementa il valore del semaforo il cui indice e` SEM; se il nuovo valore
//  non e` positivo sblocca un processo bloccato su SEM
//  
extern "C" void sem_signal(int sem);

// Alloca DIM bytes nella memoria dinamica
//
extern "C" void *mem_alloc(int dim);

// Rilascia una zona di memoria indirizzata da PV precedentemente allocata
//  con mem_alloc()
//
extern "C" void mem_free(void *pv);

// Attende N interruzioni del timer di sistema
//
extern "C" void delay(int n);

// Legge QUANTI caratteri dalla interfaccia seriale SERIAL, scrivendoli in
//  QUANTI, riportando eventuali errori in ERRORE
//
extern "C" void readse_n(int serial, char vetti[], int quanti, char &errore);

// Legge al massimo QUANTI caratteri dalla interfaccia seriale SERIAL,
//  smette quando ne trova uno nullo. Mette il risultato in vetti e
//  lascia il numero di caratteri letti in QUANTI ed un eventuale codice
//  di errore in ERRORE
//
extern "C" void readse_ln(int serial, char vetti[], int &quanti, char &errore);

// Preleva al massimo QUANTI caratteri da VETTO fino ad incontrarne uno
//  nullo, scrivendoli sull' interfaccia seriale SERIAL, lasciando il
//  numero di caratteri letti in QUANTI.
//
extern "C" void writese_0(int serial, char vetto[], int &quanti);

// Preleva QUANTI caratteri da VETTO scrivendoli sull' interfaccia
//  seriale SERIAL
//
extern "C" void writese_n(int serial, char vetto[], int quanti);

// Legge QUANTI caratteri dal terminale TERM e li mette in VETTI
//
extern "C" void term_read_n(int term, char vetti[], int &quanti);

// Scrive QUANTI caratteri sul terminale TERM, prelevandoli da VETTI
//
extern "C" void term_write_n(int term, char vetti[], int quanti);

// Costanti
//
const int LIV_SISTEMA = 0, LIV_UTENTE = 3;

#endif

